#ifndef _BUDGIE_CONFIG_H_
#define _BUDGIE_CONFIG_H_

/* i.e. /usr/share/[daemonname] */
extern const char *BUDGIE_EXTRAS_DATADIR;

/* i.e. /etc/[daemonname] */
extern const char *BUDGIE_EXTRAS_SYSCONFDIR;

/* i.e. daemonname */
extern const char *BUDGIE_EXTRAS_DAEMONNAME;

#endif

